//==================================================================
//
//	File name	:	net.js
//
//	Summary		:	Ajax, Download hole classes
//
//	Ceated		:	2018/12/06
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (net) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	net.METHOD_GET = 0;
	net.METHOD_POST = 1;
	net.STATUS_READY = 0;
	net.STATUS_REQUEST = 1;
	net.STATUS_RESPONSE = 2;
	net.STATUS_TIMEOUT = 3;
	net.STATUS_ABORT = 4;
	net.STATUS_ERROR = 9;
	net.RESPONSE_TEXT = 0;
	net.RESPONSE_DOCUMENT = 1;
	net.RESPONSE_ARRAYBUFFER = 2;
	net.RESPONSE_BLOB = 3;
	net.RESPONSE_JSON = 4;
	
	// Grobal variables
	net.ajaxes = {};
	
	
	//==========================================================
	//
	//	Class		:	Ajax
	//
	//	Summary		:	Class of Ajax
	//
	//==========================================================
	net.Ajax = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					sURL,			// * Target URL[string]
					iMethod,		// * Method[int]
					iTimeout,		// Timeout[int]
					fpOk,			// Function on OK[function]
					fpNg,			// Function on NG[function]
					fpTimeout,		// Function on Timeout[function]
					fpAbort,		// Function on Abort[function]
					fpProgress,		// Function on Progress[function]
					bMultipart,		// Multipart or not[boolean]
					fpUpProgress)	// Function on Upload Progress[function]
		{
			// Check parameters
			if(typeof sID != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ID,
											"net.js", "jaswin.net.Ajax.constructor");
			
			if(net.ajaxes[sID])
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_DUPLICATION_ID, sID),
											"net.js", "jaswin.net.Ajax.constructor");
			
			if(typeof sURL != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_URL, sID),
											"net.js", "jaswin.net.Ajax.constructor");
			
			if(typeof iMethod != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_METHOD, sID),
											"net.js", "jaswin.net.Ajax.constructor");
			
			if((iMethod != net.METHOD_GET) && (iMethod != net.METHOD_POST))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_METHOD, sID),
											"net.js", "jaswin.net.Ajax.constructor");
			
			if(iTimeout)
			{
				if(typeof iTimeout != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TIMEOUT, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpOk)
			{
				if(typeof fpOk != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_OK, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpNg)
			{
				if(typeof fpNg != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_NG, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpTimeout)
			{
				if(typeof fpTimeout != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_TIMEOUT, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpAbort)
			{
				if(typeof fpAbort != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_ABORT, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpProgress)
			{
				if(typeof fpProgress != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_PROGRESS, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if((bMultipart) && (iMethod == net.METHOD_GET))
			{
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_METHOD, sID),
											"net.js", "jaswin.net.Ajax.constructor");
			}
			
			if(fpUpProgress)
			{
				if(typeof fpUpProgress != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_UPLOAD_PROGRESS, sID),
												"net.js", "jaswin.net.Ajax.constructor");
			}
			
			// Set properties
			this.id = sID;							// ID[String]
			this.url = sURL;						// Target URL[string]
			this.method = iMethod;					// Method[int]
			this.contentType = null;				// Content type header[string]
			this.data = null;						// Data[object]
			this.timeout = iTimeout;				// Timeout[int]
			this.onOk = fpOk;						// Function on OK[function]
			this.onNG = fpNg;						// Function on NG[function]
			this.onTimeout = fpTimeout;				// Function on Timeout[function]
			this.onAbort = fpAbort;					// Function on Abort[function]
			this.onProgress = fpProgress;			// Function on Progress[function]
			this.onUpProgress = fpUpProgress;		// Function on Upload Progress[function]
			this.httpRequest = null;				// Request object[XMLHttpRequest]
			this.status = net.STATUS_READY;			// Status[int]
			this.loaded = 0;						// Loaded bytes[int]
			this.total = 0;							// Total bytes[int]
			this.responseType = net.RESPONSE_TEXT;	// Response type[int]
			this.multipart = false;					// Multipart or not[boolean]
			
			if(bMultipart)
			{
				this.data = new FormData();
				this.multipart = true;
				this.contentType = "multipart/form-data";
			}
			
			net.ajaxes[this.id] = this;
		}
		
		//==================================================
		//	Delete this
		//==================================================
		destroy()
		{
			delete net.ajaxes[this.id];
		}
		
		//==================================================
		//	Set URL
		//==================================================
		setUrl(sURL)
		{
			if(typeof sURL != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_URL, this.id),
											"net.js", "jaswin.net.Ajax.setUrl");
			
			this.url = sURL;
		}
		
		//==================================================
		//	Set method
		//==================================================
		setMethod(iMethod)
		{
			if(typeof iMethod != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_METHOD, this.id),
											"net.js", "jaswin.net.Ajax.setMethod");
			
			if((iMethod != net.METHOD_GET) && (iMethod != net.METHOD_POST))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_METHOD, this.id),
											"net.js", "jaswin.net.Ajax.setMethod");
			
			if((typeof this.data == "object")
			&& (this.multipart)
			&& (iMethod != net.METHOD_POST))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_METHOD, this.id),
											"net.js", "jaswin.net.Ajax.setMethod");
			
			this.method = iMethod;
		}
		
		//==================================================
		//	Set timeout
		//==================================================
		setTimeout(iTimeout)
		{
			if(typeof iTimeout != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TIMEOUT, this.id),
											"net.js", "jaswin.net.Ajax.setTimeout");
			
			this.timeout = iTimeout;
		}
		
		//==================================================
		//	Set OK event
		//==================================================
		setOkEvent(fpOk)
		{
			if(typeof fpOk != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_OK, this.id),
											"net.js", "jaswin.net.Ajax.setOkEvent");
			this.onOk = fpOk;
		}
		
		//==================================================
		//	Set NG event
		//==================================================
		setNgEvent(fpNg)
		{
			if(typeof fpNg != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_NG, this.id),
											"net.js", "jaswin.net.Ajax.setNgEvent");
			this.onNG = fpNg;
		}
		
		//==================================================
		//	Set Timeout event
		//==================================================
		setTimeoutEvent(fpTimeout)
		{
			if(typeof fpTimeout != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_TIMEOUT, this.id),
											"net.js", "jaswin.net.Ajax.setTimeoutEvent");
			this.onTimeout = fpTimeout;
		}
		
		//==================================================
		//	Set Abort event
		//==================================================
		setAbortEvent(fpAbort)
		{
			if(typeof fpAbort != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_ABORT, this.id),
											"net.js", "jaswin.net.Ajax.setAbortEvent");
			this.onAbort = fpAbort;
		}
		
		//==================================================
		//	Set Progress event
		//==================================================
		setProgressEvent(fpProgress)
		{
			if(typeof fpProgress != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_PROGRESS, this.id),
											"net.js", "jaswin.net.Ajax.setProgressEvent");
			this.onProgress = fpProgress;
		}
		
		//==================================================
		//	Set Upload progress event
		//==================================================
		setUploadProgressEvent(fpUpProgress)
		{
			if(typeof fpUpProgress != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_UPLOAD_PROGRESS, this.id),
											"net.js", "jaswin.net.Ajax.setUploadProgressEvent");
			this.onUpProgress = fpUpProgress;
		}
		
		//==================================================
		//	Set data
		//==================================================
		setData(sContentType, oData)
		{
			if(typeof sContentType != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTENT_TYPE, this.id),
											"net.js", "jaswin.net.Ajax.setData");
			
			if(this.multipart)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_CANNOT_IF_MULTIPART, this.id),
											"net.js", "jaswin.net.Ajax.setData");
			
			if((sContentType.toUpperCase() != "APPLICATION/X-WWW-FORM-URLENCODED")
			&& (this.method != net.METHOD_POST))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_METHOD, this.id),
											"net.js", "jaswin.net.Ajax.setData");
			
			this.contentType = sContentType;
			this.data = oData;
		}
		
		//==================================================
		//	Set parameter
		//==================================================
		setParams(saKey, oaVal)
		{
			if(!Array.isArray(saKey))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_KEY, this.id),
											"net.js", "jaswin.net.Ajax.setParams");
			
			if(!Array.isArray(oaVal))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"net.js", "jaswin.net.Ajax.setParams");
			
			if(saKey.length != oaVal.length)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_NUMBER_DATA, this.id),
											"net.js", "jaswin.net.Ajax.setParams");
			
			let iCounter;
			let sData = "";
			
			for(iCounter = 0; iCounter < saKey.length; iCounter++)
			{
				if(typeof saKey[iCounter] != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_KEY, this.id),
												"net.js", "jaswin.net.Ajax.setParams");
				
				if((!this.contentType) || (this.contentType.toUpperCase() != "MULTIPART/FORM-DATA"))
				{
					if(typeof oaVal[iCounter] != "string")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
													"net.js", "jaswin.net.Ajax.setParams");
				}
				
				if(this.multipart)
					this.data.append(saKey[iCounter], oaVal[iCounter]);
				else
				{
					if(iCounter != 0)
						sData += "&";
					sData += encodeURIComponent(saKey[iCounter]) + "=" + encodeURIComponent(oaVal[iCounter]);
				}
			}
			
			if((!this.contentType) || (!this.multipart))
			{
				this.contentType = "application/x-www-form-urlencoded";
				this.data = sData;
			}
		}
		
		//==================================================
		//	Add parameter
		//==================================================
		addParam(sKey, oVal)
		{
			if(typeof sKey != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_KEY, this.id),
											"net.js", "jaswin.net.Ajax.addParam");
			
			if((!this.contentType) || (this.contentType.toUpperCase() != "MULTIPART/FORM-DATA"))
			{
				if(typeof oVal != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
												"net.js", "jaswin.net.Ajax.addParam");
			}
			
			if(this.contentType)
			{
				if((this.contentType.toUpperCase() != "APPLICATION/X-WWW-FORM-URLENCODED")
				&& (this.contentType.toUpperCase() != "MULTIPART/FORM-DATA"))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_CONTENT_TYPE, this.id),
												"net.js", "jaswin.net.Ajax.addParam");
			}
			else
				this.contentType = "application/x-www-form-urlencoded";
			
			if(this.multipart)
			{
				if(!this.data)
					this.data = new FormData();
				this.data.append(sKey, oVal);
			}
			else
			{
				if(this.data == null)
					this.data = "";
				
				if(this.data.length > 0)
					this.data += "&";
				
				this.data += encodeURIComponent(sKey) + "=" + encodeURIComponent(oVal);
			}
		}
		
		//==================================================
		//	Send request
		//==================================================
		request(iResponseType)
		{
			if(iResponseType)
			{
				if((iResponseType != net.RESPONSE_TEXT)
				&& (iResponseType != net.RESPONSE_DOCUMENT)
				&& (iResponseType != net.RESPONSE_ARRAYBUFFER)
				&& (iResponseType != net.RESPONSE_BLOB)
				&& (iResponseType != net.RESPONSE_JSON))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_RESPONSE_TYPE, this.id),
												"net.js", "jaswin.net.Ajax.request");
				
				this.responseType = iResponseType;
			}
			
			this.status = net.STATUS_READY;
			this.httpRequest = new XMLHttpRequest();
			let oSendData;
			
			if(this.timeout)
			{
				// Time out setting
				this.httpRequest.timeout = this.timeout * 1000;
				this.httpRequest.addEventListener("timeout", this.timeoutEvent = (e) => {
					this.status = net.STATUS_TIMEOUT;
				}, false);
			}
			
			if(this.onAbort)
			{
				// Abort event
				this.httpRequest.addEventListener("abort", this.abortEvent = (e) => {
					this.status = net.STATUS_ABORT;
				}, false);
			}
			
			if(this.onProgress)
			{
				// Progress event
				this.httpRequest.addEventListener("progress", this.progressEvent = (e) => {
					this.progress(e.loaded, e.total);
				}, false);
			}
			
			if((this.multipart) && (this.onUpProgress))
			{
				// Upload progress event
				this.httpRequest.upload.addEventListener("progress", this.upProgressEvent = (e) => {
					this.progress(e.loaded, e.total);
				}, false);
			}
			
			if(this.method == net.METHOD_GET)
			{
				// GET method
				let sURL = this.url;
				
				if((this.data) && (this.data.length > 0))
					sURL + "?" + this.data;
				
				this.httpRequest.open("GET", sURL, true);
				if(this.responseType)
				{
					if(this.responseType == net.RESPONSE_TEXT)
						this.httpRequest.responseType = "text";
					else if(this.responseType == net.RESPONSE_DOCUMENT)
						this.httpRequest.responseType = "document";
					else if(this.responseType == net.RESPONSE_ARRAYBUFFER)
						this.httpRequest.responseType = "arraybuffer";
					else if(this.responseType == net.RESPONSE_BLOB)
						this.httpRequest.responseType = "blob";
					else if(this.responseType == net.RESPONSE_JSON)
					{
						if(jaswin.BROWSER == "ie")
							this.httpRequest.responseType = "text";
						else
							this.httpRequest.responseType = "json";
					}
				}
				this.httpRequest.setRequestHeader("Pragma", "no-cache");
				this.httpRequest.setRequestHeader("Cache-Control", "no-cache");
				oSendData = null;
			}
			else
			{
				// POST method
				this.httpRequest.open("POST", this.url, true);
				if(this.responseType)
				{
					if(this.responseType == net.RESPONSE_TEXT)
						this.httpRequest.responseType = "text";
					else if(this.responseType == net.RESPONSE_DOCUMENT)
						this.httpRequest.responseType = "document";
					else if(this.responseType == net.RESPONSE_ARRAYBUFFER)
						this.httpRequest.responseType = "arraybuffer";
					else if(this.responseType == net.RESPONSE_BLOB)
						this.httpRequest.responseType = "blob";
					else if(this.responseType == net.RESPONSE_JSON)
					{
						if(jaswin.BROWSER == "ie")
							this.httpRequest.responseType = "text";
						else
							this.httpRequest.responseType = "json";
					}
				}
				this.httpRequest.setRequestHeader("Content-Type", this.contentType);
				oSendData = this.data;
			}
			
			this.httpRequest.addEventListener("loadend", this.loadEndEvent = (e) => {
				this.response();
			}, false);
			
			// Send data
			this.httpRequest.send(oSendData);
			this.status = net.STATUS_REQUEST;
		}
		
		//==================================================
		//	Abort request
		//==================================================
		abort()
		{
			if(this.httpRequest)
				this.httpRequest.abort();
		}
		
		//==================================================
		//	Progress
		//==================================================
		progress(iLoaded, iTotal)
		{
			this.loaded = iLoaded;
			this.total = iTotal;
			
			if(this.onProgress)
			{
				this.onProgress(this);
			}
			else if(this.onUpProgress)
			{
				this.onUpProgress(this);
			}
		}
		
		//==================================================
		//	Receive response
		//==================================================
		response()
		{
			if((this.httpRequest.status == 200)
			|| (this.httpRequest.status == 304))
			{
				// Response OK(200, 304)
				this.status = net.STATUS_RESPONSE;
				if(this.onOk)
					this.onOk(this);
			}
			else if(this.status == net.STATUS_TIMEOUT)
			{
				// Timeout
				if(this.onTimeout)
					this.onTimeout(this);
				else if(this.onNG)
					this.onNG(this);
			}
			else if(this.status == net.STATUS_ABORT)
			{
				// Abort
				if(this.onAbort)
					this.onAbort(this);
			}
			else
			{
				// Response NG(404, 500, ...)
				this.status = net.STATUS_ERROR;
				if(this.onNG)
					this.onNG(this);
			}
		}
		
		//==================================================
		//	Get HTTP status
		//==================================================
		getHttpStatus()
		{
			if(!this.httpRequest)
				return null;
			
			return this.httpRequest.status;
		}
		
		//==================================================
		//	Get returned response
		//==================================================
		getResponse()
		{
			if(!this.httpRequest.response)
				return null;
			
			return this.httpRequest.response;
		}
		
		//==================================================
		//	Get returned response by text
		//==================================================
		responseText()
		{
			if(!this.httpRequest)
				return null;
			
			if(this.responseType != net.RESPONSE_TEXT)
			{
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_RESPONSE_TYPE, this.id),
												"net.js", "jaswin.net.Ajax.responseText");
			}
			
			return this.httpRequest.responseText;
		}
		
		//==================================================
		//	Get returned response by XML
		//==================================================
		responseXML()
		{
			if(!this.httpRequest)
				return null;
			
			if(this.responseType != net.RESPONSE_TEXT)
			{
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_RESPONSE_TYPE, this.id),
												"net.js", "jaswin.net.Ajax.responseXML");
			}
			
			return this.httpRequest.responseXML;
		}
		
		//==================================================
		//	Show save file dialog
		//==================================================
		showSaveDialog(sFileName)
		{
			if(!this.httpRequest.response)
				return;
			
			if((!this.responseType)
			|| (this.responseType != net.RESPONSE_BLOB))
			{
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_RESPONSE_TYPE, this.id),
												"net.js", "jaswin.net.Ajax.showSaveDialog");
			}
			
			if(typeof sFileName != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_FILE_NAME, this.id),
												"net.js", "jaswin.net.Ajax.showSaveDialog");
			
			let oContents = this.httpRequest.response;
			let oLink = document.createElement("a");
			let oUrl = window.URL.createObjectURL(new Blob([oContents], {type: oContents.type}));
			
			document.body.appendChild(oLink);
			oLink.style.display = "none";
			oLink.href = oUrl;
			oLink.download = sFileName;
			oLink.click();
			document.body.removeChild(oLink);
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	FileUpload
	//
	//	Summary		:	Class of File upload
	//
	//==========================================================
	net.FileUpload = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					oRect,			// * Size and position[FixedRect]
					bEnabled,		// * Enabled or not[boolean]
					bMultiple,		// * Multi files or not[boolean]
					sMimeType,		// MIME type[string]
					sTooltip,		// Tool tip text[string]
					iTabIndex)		// Tab index[int]
		{
			// Base class
			super(false, sID, false, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof bMultiple != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MULTIPLE, sID),
											"net.js", "jaswin.net.FileUpload.constructor");
			
			if(sMimeType)
			{
				if(typeof sMimeType != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_MIME_TYPE, sID),
												"net.js", "jaswin.net.FileUpload.constructor");
			}
			
			if(sTooltip)
			{
				if(typeof sTooltip != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, sID),
												"net.js", "jaswin.net.FileUpload.constructor");
			}
			
			if(!jaswin.button)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "button.js"),
											"net.js", "jaswin.net.FileUpload.constructor");
			
			// Set properties
			this.content = null;			// Content element for file selection[DOM Element]
			this.fileElement = null;		// File element[DOM Element]
			this.fileName = null;			// File name element[DOM Element]
			this.button = null;				// Upload or selection button[Button class]
			this.tabIndex = -1;				// Tab index[int]
			
			// Create element
			this.element = document.createElement("div");
			this.element.style.position = "absolute";
			if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
				this.element.style.right = this.rect.posX + "px";
			else
				this.element.style.left = this.rect.posX + "px";
			
			if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
				this.element.style.bottom = this.rect.posY + "px";
			else
				this.element.style.top = this.rect.posY + "px";
			
			this.element.style.width = this.rect.width + "px";
			this.element.style.height = "20px";
			this.focusElement = this.element;
			
			// File selection part
			this.content = document.createElement("div");
			this.content.className = "file-upload";
			if(!this.enabled)
				this.content.style.cursor = "default";
			this.element.appendChild(this.content);
			
			this.fileElement = document.createElement("input");
			this.fileElement.type = "file";
			this.fileElement.multiple = bMultiple;
			this.fileElement.style.width = "0px";
			this.fileElement.style.height = "0px";
			if(sMimeType)
				this.fileElement.accept = sMimeType;
			this.content.appendChild(this.fileElement);
			
			// Selected file names
			this.fileName = document.createElement("div");
			if(this.enabled)
				this.fileName.className = "file-upload-filename";
			else
				this.fileName.className = "file-upload-filename-disabled";
			this.content.appendChild(this.fileName);
			
			// Selection button
			let oButtonRect = new jaswin.FixedRect(jaswin.ANCHOR_RIGHT, jaswin.ANCHOR_TOP,
													0, 0, 20, 20);
			this.button = new jaswin.button.Button("Btn_" + this.id, oButtonRect,
													jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_ICON,
													true,
													"icon-open", null,
													-1, clickSelectionEvent);
			this.button.element.style.borderRadius = "0px 2px 2px 0px";
			this.element.appendChild(this.button.element);
			
			// Tooltip
			if((sTooltip) && (sTooltip.length > 0))
				this.element.title = sTooltip;
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.content.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
					this.setHover();
				}, false);
				this.content.addEventListener("mouseout", this.mouseOutEvent = (e) => {
					this.setOut();
				}, false);
				this.content.addEventListener("click", this.clickEvent = (e) => {
					this.click();
				}, false);
				this.element.addEventListener("keyup", this.keyUpEvent = (e) => {
					if((e.keyCode == 13) || (e.keyCode == 32))
						this.click();
				}, false);
				this.fileElement.addEventListener("change", this.changeEvent = (e) => {
					this.change();
				}, false);
			}
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			this.button.destroy();
			super.destroy();
		}
		
		//==================================================
		//	Set parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			this.button.setParent(this);
		}
		
		//==================================================
		//	Enable this file upload
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				this.content.style.cursor = "pointer";
				this.fileName.className = "file-upload-filename";
				this.button.enable(bParent);
				
				if(!this.mouseMoveEvent)
				{
					this.content.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
						this.setHover();
					}, false);
				}
				if(!this.mouseOutEvent)
				{
					this.content.addEventListener("mouseout", this.mouseOutEvent = (e) => {
						this.setOut();
					}, false);
				}
				if(!this.clickEvent)
				{
					this.content.addEventListener("click", this.clickEvent = (e) => {
						this.click();
					}, false);
				}
				if(!this.keyUpEvent)
				{
					this.element.addEventListener("keyup", this.keyUpEvent = (e) => {
						if((e.keyCode == 13) || (e.keyCode == 32))
							this.click();
					}, false);
				}
				if(!this.changeEvent)
				{
					this.fileElement.addEventListener("change", this.changeEvent = (e) => {
						this.change();
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this file upload
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.content.style.cursor = "default";
			this.fileName.className = "file-upload-filename-disabled";
			this.button.disable(bParent);
			
			if(this.mouseMoveEvent)
			{
				this.content.removeEventListener("mousemove", this.mouseMoveEvent, false);
				this.mouseMoveEvent = null;
			}
			if(this.mouseOutEvent)
			{
				this.content.removeEventListener("mouseout", this.mouseOutEvent, false);
				this.mouseOutEvent = null;
			}
			if(this.clickEvent)
			{
				this.content.removeEventListener("click", this.clickEvent, false);
				this.clickEvent = null;
			}
			if(this.keyUpEvent)
			{
				this.element.removeEventListener("keyup", this.keyUpEvent, false);
				this.keyUpEvent = null;
			}
			if(this.changeEvent)
			{
				this.fileElement.removeEventListener("change", this.changeEvent, false);
				this.changeEvent = null;
			}
		}
		
		//==================================================
		//	Set this file upload hover
		//==================================================
		setHover()
		{
			this.content.className = "file-upload-focus";
		}
		
		//==================================================
		//	Set this file upload mouse out
		//==================================================
		setOut()
		{
			this.content.className = "file-upload";
		}
		
		//==================================================
		//	Click this file upload
		//==================================================
		click()
		{
			if(!this.enabled)
				return;
			
			this.fileElement.click();
		}
		
		//==================================================
		//	Change this file upload
		//==================================================
		change()
		{
			let sFileName = "";
			let iCounter;
			
			for(iCounter = 0; iCounter < this.fileElement.files.length; iCounter++)
			{
				if(iCounter > 0)
					sFileName += ", ";
				sFileName += this.fileElement.files[iCounter].name;
			}
			
			this.fileName.innerHTML = sFileName;
		}
		
		//==================================================
		//	Get selected files
		//==================================================
		getFiles()
		{
			return this.fileElement.files;
		}
		
		//==================================================
		//	Clear contents
		//==================================================
		clear()
		{
			this.fileElement.value = "";
			this.fileName.innerHTML = "";
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	clickSelectionEvent
	//
	//	Summary		:	Event handler of click selection button
	//
	//==========================================================
	function clickSelectionEvent(oButton)
	{
		oButton.parent.click();
	}
	
}(jaswin.net = jaswin.net || {}));
